#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <string.h>

#include <glib.h>

#ifdef QUOTE_TEST
# define _
#else
# include "support.h"
#endif


/* from spawn */
#ifdef G_OS_WIN32
static void spawn_append_argument(GString *command, const char *text);
#endif


gchar *utils_quote_natively(const gchar *text)
{
#ifdef G_OS_WIN32
	GString *string = g_string_sized_new(strlen(text));

	spawn_append_argument(string, text);
	return g_string_free(string, FALSE);
#else
	return g_shell_quote(text);
#endif
}


typedef struct _Placeholder
{
	gchar c;           /* PH character, '\0' == end of PH array */
	gchar *rep;        /* PH string, NULL == "", g_free()-able */
} Placeholder;


/* the real placeholder replacement function */
static gchar *replace_holders_term(const gchar *format, const Placeholder *placeholders,
	gchar term, const gchar **end, GError **error)
{
	GString *string = g_string_sized_new(strlen(format));

	while (*format)
	{
		if (*format != '%')
			g_string_append_c(string, *format);
		else
		{
			switch (*++format)
			{
				case '%' : g_string_append_c(string, '%'); break;
				case '\0' : g_string_append_c(string, '%'); goto leave;
				case '"' :
				{
					if (term == *format)
					{
						*end = format;
						return g_string_free(string, FALSE);
					}
					else
					{
						gchar *substr = replace_holders_term(format + 1,
							placeholders, *format, &format, error);
						gchar *quoted;

						if (!substr)
							return g_string_free(string, TRUE);

						quoted = utils_quote_natively(substr);
						g_free(substr);
						g_string_append(string, quoted);
						g_free(quoted);
						break;
					}
				}
				default :
				{
					const Placeholder *ph;

					for (ph = placeholders; ph->c; ph++)
						if (ph->c == *format)
							break;

					if (!ph->c)
						g_string_append_c(string, *format);
					else if (ph->rep)
						g_string_append(string, ph->rep);
				}
			}
		}

		format++;
	}

leave:
	if (term)
	{
		g_set_error(error, G_SHELL_ERROR, G_SHELL_ERROR_BAD_QUOTING,
			_("No closing %%%c in '%s'"), term, format);
	}

	return g_string_free(string, term);
}


/* the public placeholder replacement function */
static gchar *replace_holders(const gchar *format, const Placeholder *placeholders,
	GError **error)
{
	g_return_val_if_fail(format != NULL && placeholders != NULL, FALSE);
	return replace_holders_term(format, placeholders, '\0', NULL, error);
}


/* test functions */
#include <stdio.h>

static gboolean read_line(const char *prompt, char *buffer, size_t size)
{
	fputs(prompt, stderr);
	*buffer = '\0';

	if (fgets(buffer, size, stdin))
	{
		char *s = strchr(buffer, '\n');

		if (s)
			*s = '\0';
	}

	return *buffer;
}


int main(void)
{
	static const Placeholder sample_ph[] =
	{
		{ 'e', "executable" },
		{ 'E', "exe'cut'able" },
		{ 'd', "directory\\" },
		{ 'D', "dir ect ory" },
		{ 'f', "file%% name" },
		{ 'F', "*nix \"file%%\" name" },
		{ 't', "text \"with\" quotes 'and' spaces" },
		{ 'v', NULL },
		{ 'V', "" },
		{ '\0', NULL }
	};

	const Placeholder *ph;
	char format[0x100];

	for (ph = sample_ph; ph->c; ph++)
		printf("%%%c -> %s\n", ph->c, ph->rep && *ph->rep ? ph->rep : "(empty text)");
	putchar('\n');


	while (read_line("original text: ", format, sizeof format))
	{
		GError *error = NULL;
		gchar *result = replace_holders(format, sample_ph, &error);

		if (result)
		{
			printf("expanded text = %s\n", result);
			g_free(result);
		}
		else
		{
			fprintf(stderr, "error: %s\n", error->message);
			g_error_free(error);
		}
	}

	return 0;
}


/* copied from spawn */
#ifdef G_OS_WIN32
#include <ctype.h>  /* isspace() */

static void spawn_append_argument(GString *command, const char *text)
{
	const char *s;

	if (command->len)
		g_string_append_c(command, ' ');

	for (s = text; *s; s++)
	{
		/* g_ascii_isspace() fails for '\v', and locale spaces (if any) will do no harm */
		if (*s == '"' || isspace(*s))
			break;
	}

	if (*text && !*s)
		g_string_append(command, text);
	else
	{
		g_string_append_c(command, '"');

		for (s = text; *s; s++)
		{
			const char *slash;

			for (slash = s; *slash == '\\'; slash++);

			if (slash > s)
			{
				g_string_append_len(command, s, slash - s);

				if (!*slash || *slash == '"')
				{
					g_string_append_len(command, s, slash - s);

					if (!*slash)
						break;
				}

				s = slash;
			}

			if (*s == '"')
				g_string_append_c(command, '\\');

			g_string_append_c(command, *s);
		}

		g_string_append_c(command, '"');
	}
}
#endif
